"""Box that can be dragged, changing the position of it's parent as well."""

import logging
from dataclasses import dataclass, replace
from typing import Optional, Callable, Sequence

import cocos
from shimmer.display.alignment import CenterCenter
from shimmer.display.components.box import Box, BoxDefinition
from shimmer.display.components.mouse_box import (
    MouseBox,
    MouseBoxDefinition,
    EVENT_HANDLED,
)

log = logging.getLogger(__name__)


@dataclass(frozen=True)
class SnapBoxDefinition(BoxDefinition):
    """
    Definition of a SnapBox.

    :param can_receive: Callback to test whether a DraggableBox should be allowed to snap
        to this SnapBox. Return True if yes, otherwise return False if not allowed.
    :param on_receive: Callback called when a DraggableBox snaps onto this SnapBox.
    :param on_release: Callback called when a DraggableBox snaps off this SnapBox.
    """

    can_receive: Optional[Callable[["DraggableBox"], bool]] = None
    on_receive: Optional[Callable[["DraggableBox"], None]] = None
    on_release: Optional[Callable[["DraggableBox"], None]] = None


class SnapBox(Box):
    """
    A Box that a DraggableBox can snap to.

    This causes the DraggableBox to be centered on this SnapBox when dragged over it.

    SnapBoxes are single occupancy - multiple DraggableBoxes cannot snap to the same SnapBox.
    """

    def __init__(self, definition: SnapBoxDefinition):
        """Create a new SnapBox."""
        super(SnapBox, self).__init__(definition)
        self.definition: SnapBoxDefinition = self.definition
        self.occupant: Optional[Box] = None

    @property
    def is_occupied(self) -> bool:
        """Return True if this SnapBox is currently occupied. Otherwise False."""
        return self.occupant is not None

    def can_receive(self, other: "DraggableBox") -> bool:
        """
        Return True if the given DraggableBox is allowed to snap to this SnapBox.

        :param other: The DraggableBox to be tested.
        """
        if self.is_occupied:
            return False
        elif self.definition.can_receive is not None:
            return self.definition.can_receive(other)
        return True

    def receive(self, other: "DraggableBox") -> None:
        """
        Called when a DraggableBox is snapped to this SnapBox.

        Calls the "on_receive" callback in the definition, if given.

        :param other: The DraggableBox that has snapped to this SnapBox.
        """
        self.occupant = other
        if self.definition.on_receive is not None:
            self.definition.on_receive(other)

    def release(self, other: "DraggableBox") -> None:
        """
        Called when a DraggableBox is no longer snapped to this SnapBox.

        Calls the "on_release" callback in the definition, if given.

        :param other: The DraggableBox that is not longer snapped to this SnapBox.
        """
        self.occupant = None
        if self.definition.on_release is not None:
            self.definition.on_release(other)


@dataclass(frozen=True)
class DraggableBoxDefinition(MouseBoxDefinition):
    """
    Definition of a draggable box.

    MouseBox actions will be overwritten when initialising the DraggableBox

    :param snap_boxes: List of SnapBoxes which the draggable anchor will snap to.
        Snapping means when the anchor is dragged over a snap box, their centers will be aligned.
    :param must_be_snapped: If True then this DraggableBox must always be snapped to a SnapBox.
        If it is dragged off a SnapBox and not onto another one then it will return to its
        previous position.
        If True, then "snap_boxes" must be given.
        Note: This is not enforced on creation; only on subsequent drags.
              Use DraggableBox.snap_to(snap_box) to set up snap behaviour on creation.
    """

    snap_boxes: Optional[Sequence[SnapBox]] = None
    must_be_snapped: bool = False

    def __post_init__(self):
        """Perform validation."""
        if self.must_be_snapped:
            if self.snap_boxes is None or len(self.snap_boxes) == 0:
                raise ValueError(
                    f"`snap_boxes` must be given if `requires_snap_box` is True. {self.snap_boxes=}"
                )


class DraggableBox(MouseBox):
    """Box that can be dragged, changing the position of it's parent as well."""

    def __init__(self, definition: DraggableBoxDefinition):
        """
        Creates a new DraggableBox.

        :param definition: DraggableBoxDefinition defining the shape of the draggable box.
        """
        defn = replace(
            definition,
            on_press=self.start_dragging,
            on_release=self.stop_dragging,
            on_drag=self.handle_drag,
        )
        super(DraggableBox, self).__init__(defn)
        self.definition: DraggableBoxDefinition = self.definition
        self._currently_snapped_to: Optional[SnapBox] = None
        self._snap_drag_record: cocos.draw.Vector2 = cocos.draw.Vector2(0, 0)

    def _should_handle_mouse_press(self, buttons: int) -> bool:
        """Should only handle events if this box is attached to something."""
        return self.parent is not None

    def _should_handle_mouse_release(self, buttons: int) -> bool:
        """Should only handle events if this box is attached to something."""
        return self.parent is not None

    def _should_handle_mouse_drag(self) -> bool:
        """Should only handle drag is this box currently is being dragged."""
        return self._currently_dragging

    def start_dragging(
        self, box: "MouseBox", x: int, y: int, buttons: int, modifiers: int,
    ) -> bool:
        """Set this box as being dragged, and reset previous drag knowledge."""
        self._snap_drag_record = cocos.draw.Vector2(0, 0)
        return super(DraggableBox, self).start_dragging(box, x, y, buttons, modifiers)

    def stop_dragging(
        self, box: "MouseBox", x: int, y: int, buttons: int, modifiers: int,
    ) -> bool:
        """
        Set this box as no longer being dragged.

        If this DraggableBox is defined with `must_be_snapped` then move it back to the current
        SnapBox if needed.
        """
        if self.definition.must_be_snapped and self._currently_snapped_to is not None:
            self._snap_drag_record = cocos.draw.Vector2(0, 0)
            self._align_with_snap_box(self._currently_snapped_to)
        return super(DraggableBox, self).stop_dragging(box, x, y, buttons, modifiers)

    def handle_drag(
        self, box: Box, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int
    ) -> Optional[bool]:
        """
        While the mouse is pressed on the area, keep updating the position.

        If snap_boxes are defined, then this DraggableBox will center its parent (and therefore
        itself) to one of the SnapBoxes if they overlap and other conditions are met as defined
        in the SnapBoxDefinition.
        """
        if self.definition.snap_boxes is None:
            self.parent.position += cocos.draw.Vector2(dx, dy)
        else:
            self._snap_drag_record += (dx, dy)
            self.parent.position += self._snap_drag_record
            # Move the parent by the snap record so we can test if we still intersect a snap point
            # It will get moved back it we still intersect the current snap point.
            # If we don't intersect with any snap point, then we will have just moved the parent
            # by the correct displacement anyway.
            for snap_box in self.definition.snap_boxes:
                if (
                    snap_box is self._currently_snapped_to or snap_box.can_receive(self)
                ) and self.world_rect.intersects(snap_box.world_rect):
                    self.snap_to(snap_box)
                    break
            else:
                # If no longer intersecting any valid snap boxes, then stop being snapped
                self._snap_drag_record = cocos.draw.Vector2(0, 0)
                self.unsnap_if_snapped()

        return EVENT_HANDLED

    def snap_to(self, snap_box: SnapBox) -> None:
        """
        Call to snap this DraggableBox to the given SnapBox.

        This aligns the center of this box with the snap box by moving the parent of this box.

        :param snap_box: The SnapBox to snap to.
        """
        # Only release/receive to snap box if it has changed.
        if self._currently_snapped_to is not snap_box:
            if self._currently_snapped_to is not None:
                self._currently_snapped_to.release(self)

            self._currently_snapped_to = snap_box
            snap_box.receive(self)

        # Always align with the target box, regardless of whether it has changed or not.
        self._align_with_snap_box(snap_box)

    def unsnap_if_snapped(self) -> None:
        """
        Call to unsnap from the current SnapBox, if there is a current SnapBox.

        If this DraggableBox is defined as `must_be_snapped` then this has no effect.

        Does not move this DraggableBox, but notifies the current SnapBox (if there is one)
        that this box is no longer snapped to it.
        """
        # Don't unsnap if this DraggableBox must always be snapped - this essentially
        # reserves its current snap target for it to return to if the drag ends before
        # it reaches another snap target.
        if not self.definition.must_be_snapped:
            if self._currently_snapped_to is not None:
                self._currently_snapped_to.release(self)
            self._currently_snapped_to = None

    def _align_with_snap_box(self, snap_box: SnapBox) -> None:
        """Align the parent of this DraggableBox with the given snap box."""
        log.debug(f"Aligning {self} with snap box {snap_box}.")
        alignment_required = self.vector_between_anchors(
            snap_box, CenterCenter, CenterCenter
        )
        self.parent.position += alignment_required
        if self._currently_snapped_to is not None:
            # Set the drag record to be the inverse motion. This allows the user to drag on/off
            # the snap point repeatedly right at the edge of the snap box.
            self._snap_drag_record = -alignment_required
            self._currently_snapped_to = snap_box



"""Widgets for asking questions of the user."""

from dataclasses import dataclass, field
from typing import Sequence, Optional, Protocol, Set, Union, Any

from .text_box import EditableTextBoxDefinition
from ..components.box import Box


class NoArgumentsCallback(Protocol):
    """Protocol defining the signature of a callback that takes no arguments."""

    def __call__(self) -> None:
        """Function signature of a basic callback."""
        pass


class OnQuestionChangeCallback(Protocol):
    """Protocol defining the signature of a callback when a question answer changes."""

    def __call__(self, answer: Any) -> None:
        """Function signature of a basic callback."""
        pass


class MultipleChoiceSelectionChangedCallback(Protocol):
    """Protocol defining the signature of a callback when a multiple choice option is selected."""

    def __call__(
        self,
        currently_selected: Set[Union[str, Box]],
        changed_choice: Union[str, Box],
        choice_state: bool,
    ) -> None:
        """Function signature of a multiple choice response callback."""
        pass


@dataclass(frozen=True)
class QuestionDefinition:
    """Definition of a question requesting user input."""

    text: str

    # Whenever the user makes a change to the question result, this function is called.
    # This is called with the current result of the question. For example this could be:
    #    - The set of multiple choice answer selected.
    #    - The current text in a text input box.
    on_change: Optional[OnQuestionChangeCallback] = None

    # If `confirmation_required`, Ok/Cancel buttons are displayed alongside the question.
    # This function is called when the Ok button is pressed.
    on_confirm: Optional[NoArgumentsCallback] = None

    # Callback that is called if a question is cancelled.
    # For example, this could be when a Window displaying a question is closed using the X button.
    on_cancel: Optional[NoArgumentsCallback] = None

    @property
    def confirmation_required(self) -> bool:
        """
        True if the question requires confirmation, otherwise False.

        Typically, if this is True, Ok/Cancel buttons will be displayed.

        If False, then typically the question will be a multiple-choice question where
        selecting one choice is sufficient to answer the question without further confirmation.
        """
        return self.on_confirm is not None


@dataclass(frozen=True)
class MultipleChoiceQuestionDefinition(QuestionDefinition):
    """
    Definition of a question with a defined set of answers.

    Users may select multiple answers if `allow_multiple` is set to True.
    """

    # The choices to include. The order of this list is the order in which the choices will be
    # presented to the user (e.g. from left to right).
    choices: Sequence[Union[str, Box]] = field(default_factory=list)

    # Whether to allow multiple options to be selected at once.
    allow_multiple: bool = False

    # Set of choices that will be selected on creation.
    defaults: Set[Union[str, Box]] = field(default_factory=set)

    def __post_init__(self):
        """Validate the definition."""
        if not self.allow_multiple:
            if len(self.defaults) > 1:
                raise ValueError(
                    "Cannot have multiple defaults if `allow_multiple` is False."
                )
        if not self.defaults.issubset(set(self.choices)):
            invalid_defaults = self.defaults.difference(set(self.choices))
            raise ValueError(
                f"All defaults must be in choices. Invalid defaults: {invalid_defaults}"
            )


@dataclass(frozen=True)
class TextInputQuestionDefinition(QuestionDefinition):
    """Definition of a question that asks for the user to input text."""

    text_box_definition: EditableTextBoxDefinition = EditableTextBoxDefinition()

    @property
    def confirmation_required(self) -> bool:
        """Always require confirmation to indicate that text entry has completed."""
        return True



"""
A set of buttons that users can choose one, or many of.

For example, this can be used to create:
- a radio button
- a "select all that apply" set of buttons
"""

from dataclasses import replace
from typing import Any, Optional

from shimmer.display.widgets.multiple_choice_buttons import (
    MultipleChoiceButtonsDefinition,
    MultipleChoiceButtons,
)
from shimmer.display.widgets.question_definition import OnQuestionChangeCallback
from shimmer.display.widgets.window import Window
from .base_dialog_window import create_base_dialog_window


def callback_with_answer_and_close_window(
    on_change: Optional[OnQuestionChangeCallback], window: Window
) -> OnQuestionChangeCallback:
    """Create a callback that calls both the on_change and window.close methods."""

    def inner(answer: Any) -> None:
        if on_change is not None:
            on_change(answer)
        # Use kill rather than close because we don't want to call the on_cancel that using
        # the close button would do.
        window.kill()

    return inner


def create_multiple_choice_question_dialog(
    definition: MultipleChoiceButtonsDefinition,
) -> Window:
    """
    Create a window containing a multiple choice question.

    If the question allows multiple answers, a submit button will be shown that closes the window.
    The submit button doe not call any additional callback as each answer choice will have called
    back already.

    If confirmation is not required, the window will close as soon as an option is chosen.
    Otherwise, "Ok" and "Cancel" buttons will be added.

    :param definition: Definition of the question.
    """
    window = create_base_dialog_window(definition.question)

    if not definition.question.confirmation_required:
        # If confirmation is not required then close the window when any answer is selected.
        definition = replace(
            definition,
            question=replace(
                definition.question,
                on_change=callback_with_answer_and_close_window(
                    definition.question.on_change, window
                ),
            ),
        )

    buttons = MultipleChoiceButtons(definition)
    window.add_child_to_body(buttons)

    return window



"""
Module defining common dialogs.

A dialog is a Window with buttons for a multiple choice question.
For example, an "Are you sure" dialog with yes/no answers.
"""

from typing import Optional

from shimmer.display.widgets.dialogs.multiple_choice import (
    create_multiple_choice_question_dialog,
    MultipleChoiceButtonsDefinition,
)
from shimmer.display.widgets.question_definition import (
    MultipleChoiceQuestionDefinition,
    NoArgumentsCallback,
    OnQuestionChangeCallback,
)
from shimmer.display.widgets.window import Window


def create_are_you_sure_dialog(
    on_answer: OnQuestionChangeCallback,
    on_cancel: Optional[NoArgumentsCallback] = None,
) -> Window:
    """
    Create a standard "Are you sure?" dialog.

    Using the `close` button is equivalent to a `No` answer.

    :param on_answer: Method to call when a selection is made.
    :param on_cancel: Method to call if the X button is used instead of answering the question.
    :return: Window containing Yes/No buttons.
    """
    return create_multiple_choice_question_dialog(
        MultipleChoiceButtonsDefinition(
            question=MultipleChoiceQuestionDefinition(
                text="Are you sure?",
                choices=["Yes", "No"],
                on_change=on_answer,
                on_cancel=on_cancel,
                allow_multiple=False,
            )
        )
    )



"""
Module defining a Box that contains a cocos Sprite.

This allows display of static images and animated sprites.
"""

from dataclasses import dataclass
from io import IOBase
from pathlib import Path
from typing import Union, Optional

import pyglet
from pyglet.image import AbstractImage, Animation

from cocos.sprite import Sprite
from .box import Box, BoxDefinition

ANIMATION_SUFFIXES = [".gif"]


def load_image_from_path(path: Path) -> Union[AbstractImage, Animation]:
    """
    Load an image found at path into a pyglet object.

    If the image has a ".gif" extension then it will be loaded as an Animation object.

    :param path: Path to the
    :return: The pyglet object holding the loaded image or animation.
    """
    if path.suffix in ANIMATION_SUFFIXES:
        load_func = pyglet.image.load_animation
    else:
        load_func = pyglet.image.load
    with open(str(path), "rb") as fi:
        return load_func(filename=path.name, file=fi)


@dataclass(frozen=True)
class SpriteBoxDefinition(BoxDefinition):
    """
    Definition of a Box that contains a cocos Sprite.

    The texture of the Sprite is stored on this definition and therefore can be used by multiple
    instances of SpriteBox to save on resources and loading times.

    :param image: Image to display in the box. Possible types are:
        - str: name of the image resource loaded into pyglet
        - Path: path to the image file. If a relative path is given, then it is relative to the
                current working directory.
        - IOBase: Binary stream containing image data.
                  For example, from `open(image_path, "rb")`.
        - AbstractImage: An instance of a pyglet AbstractImage subclass.
        - Animation: An instance of a pyglet Animation.
    """

    image: Optional[Union[str, Path, IOBase, AbstractImage, Animation]] = None

    def __post_init__(self):
        """Handle loading the image as needed upon creation."""
        actual_image: Union[AbstractImage, Animation]

        if self.image is None:
            raise ValueError(f"`image` cannot be None for {self.__class__.__name__}.")

        if isinstance(self.image, str):
            actual_image = pyglet.resource.image(self.image)
        elif isinstance(self.image, Path):
            if self.image.is_absolute():
                actual_image = load_image_from_path(self.image)
            else:
                path = Path.cwd() / self.image
                actual_image = load_image_from_path(path)
        elif isinstance(self.image, IOBase):
            # Need a name for the image to be stored in pyglet resource as.
            # There isn't a useful name associated with a stream object, so just use its ID.
            actual_image = pyglet.image.load(str(id(self.image)), file=self.image)
        else:
            actual_image = self.image

        # Have to use __setattr__ to get around `frozen=True`.
        object.__setattr__(self, "image", actual_image)


class SpriteBox(Box):
    """A Box that contains a cocos Sprite."""

    definition_type = SpriteBoxDefinition

    def __init__(self, definition: SpriteBoxDefinition):
        """Creates a new SpriteBox."""
        super(SpriteBox, self).__init__(definition)
        self.definition: SpriteBoxDefinition = self.definition
        self.sprite = Sprite(self.definition.image, anchor=(0, 0))
        self._update_sprite_scale()
        self.add(self.sprite)

    def _update_sprite_scale(self):
        """Scale the underlying Sprite such that it fits the size of this Box."""
        if self.rect.width == 0 or self.rect.height == 0:
            self.sprite.scale_x = 1
            self.sprite.scale_y = 1
            return

        scale_x = self.rect.width / self.sprite.width
        scale_y = self.rect.height / self.sprite.height

        self.sprite.scale_x = scale_x
        self.sprite.scale_y = scale_y

    def on_size_change(self):
        """When the size of this Box changes, update the Sprite to fit the new size."""
        self._update_sprite_scale()
        super(SpriteBox, self).on_size_change()



"""Module aiding the creation of text input dialogs."""

from dataclasses import replace

from .base_dialog_window import create_base_dialog_window
from ..question_definition import TextInputQuestionDefinition
from ..text_box import EditableTextBox
from ..window import Window


def create_text_input_dialog(definition: TextInputQuestionDefinition) -> Window:
    """Create a text input dialog window."""
    window = create_base_dialog_window(definition)
    text_box_definition = replace(
        definition.text_box_definition, on_change=definition.on_change
    )
    text_box = EditableTextBox(text_box_definition)
    window.add_child_to_body(text_box)
    return window



"""Base definition and creation of dialog windows."""

import logging
from typing import Optional

from pyglet.window import key

from shimmer.display.components.box_layout import BoxRow
from shimmer.display.components.mouse_box import MouseClickEventCallable
from shimmer.display.keyboard import KeyboardHandler, KeyboardHandlerDefinition
from shimmer.display.widgets.button import ButtonDefinition, Button
from shimmer.display.widgets.question_definition import QuestionDefinition
from shimmer.display.widgets.window import Window, WindowDefinition

log = logging.getLogger(__name__)


def create_base_dialog_window(definition: QuestionDefinition) -> Window:
    """
    Create a window containing a multiple choice question.

    If the question allows multiple answers, a submit button will be shown that closes the window.
    The submit button does not call any additional callback as each answer choice will have called
    back already.

    If confirmation is not required, the window will close as soon as an option is chosen.
    Otherwise, "Ok" and "Cancel" buttons will be added.

    :param definition: Definition of the question.
    """
    on_close: Optional[MouseClickEventCallable]

    def cancel_on_close() -> MouseClickEventCallable:
        """Wrapper around the mouse event callback on close that ignores all the arguments."""

        def inner(*_, **__):
            if definition.on_cancel is not None:
                definition.on_cancel()

        return inner

    if definition.on_cancel is not None:
        on_close = cancel_on_close()
    else:
        on_close = None

    window_definition = WindowDefinition(title=definition.text, on_close=on_close,)
    window = Window(window_definition)

    if definition.confirmation_required:
        # Create "Ok" and "Cancel" buttons
        # The "Ok" button calls the `on_confirm` callback with the current selection.
        # The Cancel button is equivalent to clicking the "X".
        def on_ok(*_, **__):
            if definition.on_confirm is not None:
                definition.on_confirm()
            window.close()

        def on_cancel(*_, **__):
            # We set the window to call the cancel callback on close above,
            # so just close the window here.
            window.close()

        ok_button = Button(ButtonDefinition(text="Ok", on_press=on_ok))
        cancel_button = Button(ButtonDefinition(text="Cancel", on_press=on_cancel))
        confirmation_buttons = BoxRow(boxes=(ok_button, cancel_button))

        # Add keyboard handler to make Enter/Esc hit Ok/Cancel respectively.
        keyboard_handler_defn = KeyboardHandlerDefinition(
            focus_required=False, logging_name="dialog_enter_escape",
        )
        keyboard_handler_defn.add_keyboard_action_simple(
            key=key.ENTER, action=ok_button.on_keyboard_select_press
        )
        keyboard_handler_defn.add_keyboard_action_simple(
            key=key.NUM_ENTER, action=ok_button.on_keyboard_select_press
        )
        keyboard_handler_defn.add_keyboard_action_simple(
            key=key.ESCAPE, action=cancel_button.on_keyboard_select_press,
        )
        keyboard_handler = KeyboardHandler(keyboard_handler_defn)

        confirmation_buttons.add(keyboard_handler)

        window.add_child_to_body(confirmation_buttons)

    return window



"""Module defining windows."""

from dataclasses import dataclass, replace
from typing import Dict, Optional

from ..alignment import (
    LeftTop,
    CenterBottom,
    RightTop,
)
from ..components.box import Box, BoxDefinition
from ..components.box_layout import BoxColumn, BoxLayoutDefinition
from ..components.draggable_box import DraggableBox, DraggableBoxDefinition
from ..components.focus import make_focusable, VisualAndKeyboardFocusBox
from ..components.font import FontDefinition, Calibri
from ..components.mouse_box import (
    MouseClickEventCallable,
    MouseBox,
    MouseVoidBoxDefinition,
)
from ..data_structures import Color
from ..widgets.button import Button
from ..widgets.close_button import (
    CloseButton,
    CloseButtonDefinitionBase,
)
from ..widgets.text_box import TextBox, TextBoxDefinition


@dataclass(frozen=True)
class WindowDefinition(MouseVoidBoxDefinition):
    """
    Definition of the visual style of a Window.

    Inherits from MouseVoidBoxDefinition so that the default window swallows all mouse events
    that happen within it. Other components can be added into the window that can receive
    those events first. This prevents users clicking on a window and something underneath
    the window receiving the event accidentally.

    The `height` parameter is set on init to be the sum of the `body_height` and the
    `title_bar_height`.
    """

    # Text to display in the title bar of the window.
    title: Optional[str] = None
    title_font_definition: FontDefinition = Calibri

    # Height of the title bar. If None then it will match the height of the title font.
    # This is in addition to the the `height` of the window, which controls the height of the
    # window body.
    title_bar_height: Optional[int] = None

    # Space to leave between buttons in the title bar
    title_bar_button_spacing: int = 2
    title_bar_color: Color = Color(20, 120, 255)

    # Background color of the body of the window.
    background_color: Color = Color(20, 20, 20)

    # For dynamically sized windows (i.e. width, height = None, None) add this amount of space
    # between the children of the body and the edges of the window on all sides.
    padding: int = 15

    # Callback to call when the window is closed using the close button.
    on_close: Optional[MouseClickEventCallable] = None

    def get_title_bar_height(self) -> int:
        """
        Height of the title bar.

        This is the defined height `title_bar_height` if given,
        otherwise the height of the title font.
        """
        if self.title_bar_height is not None:
            return self.title_bar_height
        return self.title_font_definition.height


class Window(MouseBox):
    """
    A Window is a draggable, closeable Box with a title bar.

    It has an inner_box for containing children for displaying inside the window.

    All nodes added to the window are placed above this VoidBox so they receive the mouse events
    first.
    """

    def __init__(self, definition: WindowDefinition):
        """
        Create a Window.

        :param definition: Definition of the Window style.
        """
        self.definition: WindowDefinition = definition
        self.title_bar_height = self.definition.get_title_bar_height()

        super(Window, self).__init__(definition)

        self._title: Optional[TextBox] = None
        self._title_boxes: Dict[str, Box] = {}
        self._title_bar_background: Optional[Box] = None
        self.focus_box: Optional[VisualAndKeyboardFocusBox] = None

        # Add the inner box, which is the main body of the window excluding the title bar.
        self.body = BoxColumn(BoxLayoutDefinition())
        self.add(self.body)

        self.update_all()

    def make_focused(self):
        """Make this window the current focus target."""
        if self.focus_box is not None:
            self.focus_box.take_focus()

    def update_all(self):
        """Update or re-create all of the components that make up the window."""
        self._update_close_button()
        self._update_title()
        self._update_title_bar_background()
        self.update_background()
        self._update_drag_zone()
        self._update_focus_box()

    def update_rect(self):
        """
        Update the cached rect definition.

        If the rect size changes, then `on_size_change` will be called.
        """
        if self.definition.is_dynamic_sized:
            # We don't use bounding_rect_of_children here because we are rearranging those
            # children in space at the same time which gets confusing - so hard coded calculation
            # for now at least.
            children_rect = self.body.rect.copy()
            children_rect.height += self.title_bar_height
            children_rect.height += 2 * self.definition.padding
            children_rect.width += 2 * self.definition.padding

            if self._title is not None:
                children_rect.width = max(
                    children_rect.width, self.minimum_title_bar_width
                )

            self._width = children_rect.width
            self._height = children_rect.height
        else:
            self._width = self.definition.width or 0
            self._height = self.definition.height or 0

        if self._width != self._rect.width or self._height != self._rect.height:
            self._rect.set_size((self._width, self._height))
            self.on_size_change()

    def on_child_size_changed(self):
        """Called when a child of the window changes size."""
        self.update_rect()
        self.update_all()
        self.body.align_anchor_with_other_anchor(
            self, CenterBottom, spacing=(0, self.definition.padding)
        )

    @property
    def minimum_title_bar_width(self) -> int:
        """
        Minimum width, in pixels, of the window title bar.

        This is determined by the length of the title text and any title bar buttons.
        """
        width = self.close_button.rect.width + self.definition.title_bar_button_spacing

        if self._title is not None:
            width += self._title.rect.width + self._title.x

        return width

    @property
    def _title_bar_button_height(self):
        """Height of the title bar buttons, accounting for spacing."""
        return self.title_bar_height - (2 * self.definition.title_bar_button_spacing)

    @property
    def _leftmost_title_bar_button_position(self) -> int:
        """Return the leftmost edge of the buttons in the title bar."""
        return min(
            [
                self.rect.width - box.x
                for box in self._title_boxes.values()
                if isinstance(box, Button)
            ]
        )

    def _update_focus_box(self):
        """Recreate the focus box of this window, preserving current focus state."""
        was_focused = False
        if self.focus_box is not None:
            was_focused = self.focus_box.is_focused
            self.remove(self.focus_box, no_resize=True)

        self.focus_box = make_focusable(self)

        if was_focused:
            self.focus_box.take_focus()

    def _update_title(self):
        """Recreate the title."""
        if self._title is not None:
            self.remove(self._title, no_resize=True)

        if self.definition.title is None:
            return

        title_definition = TextBoxDefinition(
            text=self.definition.title, height=self.definition.title_bar_height,
        )

        self._title = TextBox(title_definition)
        self.add(self._title, no_resize=True)
        self._title.align_anchor_with_other_anchor(self, LeftTop, spacing=(10, 0))

    def _update_title_bar_background(self):
        """Redefine the background color of the title bar."""
        if self._title_bar_background is not None:
            self.remove(self._title_bar_background, no_resize=True)

        self._title_bar_background = Box(
            BoxDefinition(
                width=self.rect.width,
                height=self.title_bar_height,
                background_color=self.definition.title_bar_color,
            )
        )
        self.add(self._title_bar_background, z=-1, no_resize=True)
        self._title_bar_background.align_anchor_with_other_anchor(self, LeftTop)

    def _update_close_button(self):
        """Redefine the window close button."""
        definition = replace(
            CloseButtonDefinitionBase,
            width=self._title_bar_button_height,
            height=self._title_bar_button_height,
            on_release=self.definition.on_close,
        )
        close_button = CloseButton(definition)
        self._update_title_bar_box("close", close_button)
        self.close_button.align_anchor_with_other_anchor(
            self,
            RightTop,
            spacing=(
                -self.definition.title_bar_button_spacing,
                -self.definition.title_bar_button_spacing,
            ),
        )

    @property
    def close_button(self) -> CloseButton:
        """Get the close button of this window."""
        return self._title_boxes["close"]

    @property
    def drag_box(self) -> DraggableBox:
        """Get the DraggableBox of this window."""
        return self._title_boxes["drag"]

    def _update_drag_zone(self):
        """
        Redefine the draggable area of the title bar.

        This creates a draggable area of the window covering the entire title bar to the left of
        the leftmost title bar button.
        """
        drag_box_definition = DraggableBoxDefinition(
            width=self.rect.width - self._leftmost_title_bar_button_position,
            height=self.title_bar_height,
        )
        self._update_title_bar_box("drag", DraggableBox(drag_box_definition))
        self._title_boxes["drag"].align_anchor_with_other_anchor(
            self, LeftTop,
        )

    def _update_title_bar_box(self, name: str, box: Box) -> None:
        """
        Create or re-create a title bar Box known by the given name.

        :param name: Internal identifier of the Box.
        :param box: New Box node to place in the title bar.
        """
        if name in self._title_boxes.keys():
            self.remove(self._title_boxes[name], no_resize=True)

        self.add(box, no_resize=True)
        self._title_boxes[name] = box

    def add_child_to_body(self, child: Box) -> None:
        """
        Add a child to the body of the window.

        Children will be arranged vertically in the order they were added from bottom to top.

        For more precise control, replace the `body` with any other Box you want by removing the
        default body and adding your own.

        :param child: The child Box to add.
        """
        self.body.add(child)

    def close(self):
        """Close this window by simulating a click on the close button."""
        self.close_button.on_keyboard_select_release()
